/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.chainedplugin.hdfs.hive;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.collections.MapUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.ranger.authorization.utils.StringUtil;
import org.apache.ranger.plugin.contextenricher.RangerTagForEval;
import org.apache.ranger.plugin.model.RangerPolicy;
import org.apache.ranger.plugin.policyengine.*;
import org.apache.ranger.plugin.service.RangerBasePlugin;
import org.apache.ranger.plugin.service.RangerChainedMapperPlugin;
import org.apache.ranger.plugin.util.RangerAccessRequestUtil;
import org.apache.ranger.plugin.util.RangerPerfTracer;
import org.apache.ranger.plugin.util.RangerRMSAccessResourceImpl;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

public class RangerHdfsHiveChainedPlugin extends RangerChainedMapperPlugin {
    private static final Log LOG = LogFactory.getLog(RangerHdfsHiveChainedPlugin.class);

    private static final Log PERF_CHAINED_PLUGIN_POLICY_EVAL_LOG = RangerPerfTracer.getPerfLogger("chainedplugin.policyeval");

    private static final String PROP_HDFS_TO_HIVE_AUTHORIZE_ONLY_WITH_CHAINED_POLICIES = ".mapping.hive.authorize.with.only.chained.policies";

    private static final String PROP_HDFS_TO_HIVE_ACCESSTYPE_MAPPING_READ = ".accesstype.mapping.read";
    private static final String PROP_HDFS_TO_HIVE_ACCESSTYPE_MAPPING_WRITE = ".accesstype.mapping.write";
    private static final String PROP_HDFS_TO_HIVE_ACCESSTYPE_MAPPING_EXECUTE = ".accesstype.mapping.execute";

    private static final String CHAINED_SERVICE_TYPE     = "hive";
    private static final String HDFS_ACCESS_TYPE_READ    = "read";
    private static final String HDFS_ACCESS_TYPE_WRITE   = "write";
    private static final String HDFS_ACCESS_TYPE_EXECUTE = "execute";
    private static final String HIVE_ACCESS_TYPE_SELECT  = "select";
    private static final String HIVE_ACCESS_TYPE_UPDATE  = "update";
    //private static final String HIVE_ACCESS_TYPE_CREATE  = "create";
    //private static final String HIVE_ACCESS_TYPE_DROP    = "drop";
    private static final String HIVE_ACCESS_TYPE_ALTER   = "alter";
    //private static final String HIVE_ACCESS_TYPE_REPLADMIN   = "repladmin";
    private static final String ACCESS_TYPE_SEPARATOR    = ",";

    private final Map<String, List<String>> accessTypeMapHdfsToHive;
    private final boolean                   isAuthorizeOnlyWithChainedPolicies;
    private final List<String>              hiveServiceUserNames;

    public RangerHdfsHiveChainedPlugin(RangerBasePlugin rootPlugin, String serviceName) {
        super(rootPlugin, CHAINED_SERVICE_TYPE, serviceName);

        accessTypeMapHdfsToHive = initAccessTypesMap(rootPlugin);
        isAuthorizeOnlyWithChainedPolicies = rootPlugin.getConfig().getBoolean(rootPlugin.getConfig().getPropertyPrefix() + PROP_HDFS_TO_HIVE_AUTHORIZE_ONLY_WITH_CHAINED_POLICIES, true);

        String hiveServiceUsersPropName = plugin.getConfig().getPropertyPrefix() + ".service.names";
        String hiveServiceUsers = plugin.getConfig().get(hiveServiceUsersPropName, "hive,impala");
        hiveServiceUserNames = StringUtils.isBlank(hiveServiceUsers) ? new ArrayList<>() : StringUtil.toList(hiveServiceUsers);

        LOG.info("RangerHdfsHiveChainedPlugin(" + serviceType + ", " + serviceName + ")");
    }

    @Override
    public void init() {
        LOG.info("==> RangerHdfsHiveChainedPlugin.init(" + serviceType + ", " + serviceName + ")");

        if (getIsInited()) {
            super.init();
            if (LOG.isDebugEnabled()) {
                LOG.debug("Configuration for RangerHdfsHiveChainedPlugin is :[" + plugin.getConfig() + "]");
            }
        } else {
            LOG.error("RangerHdfsHiveChainedPlugin for service [" + serviceName + "] was not initialized correctly. It will not be operational!");
        }

        LOG.info("<== RangerHdfsHiveChainedPlugin.init(" + serviceType + ", " + serviceName + ")");
    }

    @Override
    public RangerAccessResult isAccessAllowed(RangerAccessRequest hdfsRequest) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerHdfsHiveChainedPlugin.isAccessAllowed(" + hdfsRequest + ")");
        }

        RangerAccessResult         ret             = null;

        if (getIsInited()) {

            RangerPerfTracer perf = null;

            if (RangerPerfTracer.isPerfTraceEnabled(PERF_CHAINED_PLUGIN_POLICY_EVAL_LOG)) {
                perf = RangerPerfTracer.getPerfTracer(PERF_CHAINED_PLUGIN_POLICY_EVAL_LOG, "RangerHdfsHiveChainedPlugin.isAccessAllowed(resource=" + hdfsRequest.getResource().getAsString() + ")");
            }

            List<RangerAccessResource> hiveResources = getHiveResources(hdfsRequest);
            List<String> hiveAccessTypes = CollectionUtils.isNotEmpty(hiveResources) ? getHiveAccessTypes(hdfsRequest) : null;

            if (CollectionUtils.isNotEmpty(hiveResources) && CollectionUtils.isNotEmpty(hiveAccessTypes)) {
                Set<RangerTagForEval> hdfsTags = RangerAccessRequestUtil.getRequestTagsFromContext(hdfsRequest.getContext());
                Set<RangerTagForEval> allTags = hdfsTags == null ? null : new HashSet<>(hdfsTags);

                for (RangerAccessResource hiveResource : hiveResources) {

                    Set<RangerTagForEval> tagsForResource = null;
                    boolean isRowFilterEvaluated = false;
                    boolean isDataMaskEvaluated = false;

                    hiveAccessTypes = getHiveAccessTypesForResource(hdfsRequest, hiveResource, hiveAccessTypes);
                    if (CollectionUtils.isEmpty(hiveAccessTypes)) {
                        if (LOG.isDebugEnabled()) {
                            LOG.debug("List of access-types to be checked for resource [" + hiveResource + "] is empty. Denying access!");
                        }
                        ret = null;
                    }

                    for (String hiveAccessType : hiveAccessTypes) {
                        RangerAccessRequestImpl hiveAccessRequest = new RangerAccessRequestImpl(hiveResource, hiveAccessType, hdfsRequest.getUser(), hdfsRequest.getUserGroups(), hdfsRequest.getUserRoles());

                        hiveAccessRequest.setClientIPAddress(hdfsRequest.getClientIPAddress());
                        hiveAccessRequest.setClientType(hdfsRequest.getClientType());
                        hiveAccessRequest.setClusterName(hdfsRequest.getClusterName());
                        hiveAccessRequest.setClusterType(hdfsRequest.getClusterType());
                        hiveAccessRequest.setAction(hdfsRequest.getAction());
                        hiveAccessRequest.setAccessTime(hdfsRequest.getAccessTime());
                        hiveAccessRequest.setRemoteIPAddress(hdfsRequest.getRemoteIPAddress());
                        hiveAccessRequest.setRequestData(hdfsRequest.getRequestData());
                        hiveAccessRequest.setSessionId(hdfsRequest.getSessionId());
                        hiveAccessRequest.setContext(hdfsRequest.getContext());

                        // set the scope to consider columns as well
                        hiveAccessRequest.setResourceMatchingScope(RangerAccessRequest.ResourceMatchingScope.SELF_OR_DESCENDANTS);

                        // check if row-filter is applied
                        if (!isRowFilterEvaluated) {
                            hiveAccessRequest.setAccessType(HIVE_ACCESS_TYPE_SELECT);
                            RangerAccessResult rowFilteringResult = plugin.evalRowFilterPolicies(hiveAccessRequest, null);
                            isRowFilterEvaluated = true;
                            tagsForResource = RangerAccessRequestUtil.getRequestTagsFromContext(hiveAccessRequest.getContext());

                            if (rowFilteringResult != null && rowFilteringResult.isRowFilterEnabled()) {
                                if (ret == null) {
                                    ret = new RangerAccessResult(RangerPolicy.POLICY_TYPE_ROWFILTER, plugin.getServiceName(), plugin.getServiceDef(), hiveAccessRequest);
                                }
                                ret.setIsAllowed(false);
                                ret.setPolicyId(rowFilteringResult.getPolicyId());
                                ret.setPolicyVersion(rowFilteringResult.getPolicyVersion());
                                ret.setPolicyPriority(rowFilteringResult.getPolicyPriority());
                                ret.setZoneName(rowFilteringResult.getZoneName());

                                break; // no need to evaluate further policies
                            }
                        }

                        // check if column-masking is applied
                        if (!isDataMaskEvaluated) {
                            hiveAccessRequest.setAccessType(HIVE_ACCESS_TYPE_SELECT);
                            RangerAccessResult maskingResult = plugin.evalDataMaskPolicies(hiveAccessRequest, null);
                            isDataMaskEvaluated = true;
                            tagsForResource = RangerAccessRequestUtil.getRequestTagsFromContext(hiveAccessRequest.getContext());

                            if (maskingResult != null && StringUtils.isNotEmpty(maskingResult.getMaskType())) {
                                if (ret == null) {
                                    ret = new RangerAccessResult(RangerPolicy.POLICY_TYPE_DATAMASK, plugin.getServiceName(), plugin.getServiceDef(), hiveAccessRequest);
                                }
                                ret.setIsAllowed(false);
                                ret.setPolicyId(maskingResult.getPolicyId());
                                ret.setPolicyVersion(maskingResult.getPolicyVersion());
                                ret.setPolicyPriority(maskingResult.getPolicyPriority());
                                ret.setZoneName(maskingResult.getZoneName());

                                break; // no need to evaluate further policies
                            }
                        }

                        hiveAccessRequest.setAccessType(hiveAccessType);

                        RangerAccessResult hiveAccessResult = plugin.isAccessAllowed(hiveAccessRequest);

                        tagsForResource = RangerAccessRequestUtil.getRequestTagsFromContext(hiveAccessRequest.getContext());

                        if (hiveAccessResult == null || !hiveAccessResult.getIsAccessDetermined() || hiveAccessResult.getPolicyId() == -1L) { // RANGER-3294 sets isAccessDetermined to true in certain cases.
                            continue;
                        }

                        if (hiveAccessResult.getIsAllowed()) {
                            // CDPD-17227 - This problem is essentially not solvable as RMS mappings do not keep track of the columns in the table.
                            // An approximation is as follows:
                            // Check if there is a policy allowing access for all columns. This means the policy either has '*' in column resource-spec or is a table level policy.
                            // This is still not completely correct, as there may be several Hive policies which together grant access to all columns in the table. In such case, the access will be denied.
                            // Easiest but most inefficient way to do this is to check again but with scope set to SELF

                            // Set the resource-matching scope to SELF
                            RangerAccessRequest.ResourceMatchingScope oldScope = hiveAccessRequest.getResourceMatchingScope();
                            hiveAccessRequest.setResourceMatchingScope(RangerAccessRequest.ResourceMatchingScope.SELF);

                            hiveAccessResult = plugin.isAccessAllowed(hiveAccessRequest);

                            // Reset the resource-matching scope
                            hiveAccessRequest.setResourceMatchingScope(oldScope);

                            if (hiveAccessResult == null || !hiveAccessResult.getIsAccessDetermined()) {
                                continue;
                            }
                        }

                        boolean overrideResult = ret == null || hiveAccessResult.getPolicyPriority() > ret.getPolicyPriority();

                        if (!overrideResult) {
                            // override if current result is from the same policy priority, and if denies access
                            if (hiveAccessResult.getPolicyPriority() == ret.getPolicyPriority() && !hiveAccessResult.getIsAllowed()) {
                                overrideResult = true;
                            }
                        }

                        if (overrideResult) {
                            ret = hiveAccessResult;
                        } else if (ret != null) {
                            ret.setZoneName(hiveAccessResult.getZoneName());
                        }
                    }

                    if (CollectionUtils.isNotEmpty(tagsForResource)) {
                        if (CollectionUtils.isEmpty(allTags)) {
                            allTags = tagsForResource;
                        } else {
                            allTags.addAll(tagsForResource);
                        }
                    }

                }

                if (isAuthorizeOnlyWithChainedPolicies && ret == null) {
                    ret = new RangerAccessResult(RangerPolicy.POLICY_TYPE_ACCESS, plugin.getServiceName(), plugin.getServiceDef(), hdfsRequest);
                    ret.setIsAllowed(false);
                    ret.setPolicyPriority(RangerPolicy.POLICY_PRIORITY_OVERRIDE);
                }

                RangerAccessRequestUtil.setRequestTagsInContext(hdfsRequest.getContext(), allTags);
            }

            RangerPerfTracer.logAlways(perf);

        } else {
            if (LOG.isDebugEnabled()) {
                LOG.debug("isAccessAllowed() is a not operational as the chained plugin was not correctly initialized");
            }
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerHdfsHiveChainedPlugin.isAccessAllowed(" + hdfsRequest + "): ret=" + ret);
        }

        return ret;
    }

    @Override
    public Collection<RangerAccessResult> isAccessAllowed(Collection<RangerAccessRequest> hdfsRequests) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerHdfsHiveChainedPlugin.isAccessAllowed(" + hdfsRequests + ")");
        }

        Collection<RangerAccessResult> ret = new ArrayList<>();

        for (RangerAccessRequest hdfsRequest : hdfsRequests) {
            ret.add(isAccessAllowed(hdfsRequest));
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerHdfsHiveChainedPlugin.isAccessAllowed(" + hdfsRequests + "): ret=" + ret);
        }

        return ret;
    }

    private List<RangerAccessResource> getHiveResources(RangerAccessRequest hdfsRequest) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerHdfsHiveChainedPlugin.getHiveResources(" + hdfsRequest + ")");
        }

        List<RangerAccessResource> ret = null;

        if (isNotWhitelisted(hdfsRequest)) {
            if (hdfsRequest.getResource() != null) {

                ret = mappingProvider.getMappedResources(hdfsRequest);

            } else {
                LOG.warn("Unexpected null resource found in the request, ignoring");
            }
        } else {
            if (LOG.isDebugEnabled()) {
                LOG.debug("user/groups are whitelisted :[" + hdfsRequest + "]");
            }
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerHdfsHiveChainedPlugin.getHiveResources(" + hdfsRequest + "): ret=" + ret);
        }

        return ret;
    }

    private List<String> getHiveAccessTypes(RangerAccessRequest hdfsRequest) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerHdfsHiveChainedPlugin.getHiveAccessTypes(" + hdfsRequest + ")");
        }

        List<String> ret = accessTypeMapHdfsToHive.get(hdfsRequest.getAccessType());

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerHdfsHiveChainedPlugin.getHiveAccessTypes(" + hdfsRequest + ")");
        }

        return ret;
    }

    private List<String> getHiveAccessTypesForResource(RangerAccessRequest hdfsRequest, RangerAccessResource hiveResource, List<String> hiveAccessTypes) {
        List<String> ret = hiveAccessTypes;

        String user = hdfsRequest.getUser();

        if (!hiveServiceUserNames.contains(user)) {
            if (hiveResource instanceof RangerRMSAccessResourceImpl) {
                Map<String, String> additionalInfo = ((RangerRMSAccessResourceImpl) hiveResource).getAdditionalInfo();
                if (MapUtils.isNotEmpty(additionalInfo)) {
                    String tableType = additionalInfo.get("tableType");
                    if (!StringUtils.equals(tableType, "EXTERNAL_TABLE")) {
                        // Deny write access
                        if (hdfsRequest.getAccessType().equals(HDFS_ACCESS_TYPE_WRITE)) {
                            LOG.error("User [" + user + "] is allowed only [" + HDFS_ACCESS_TYPE_READ + " or " + HDFS_ACCESS_TYPE_EXECUTE + "] access to " + hdfsRequest.getResource().getAsString() + " which maps to table of type [" + tableType + "]. Access is not allowed.]");
                            ret = Collections.emptyList();
                        }
                    }
                }
            }
        }
        return ret;
    }

    private static Map<String, List<String>> initAccessTypesMap(RangerBasePlugin rootPlugin) {
        Map<String, List<String>> ret = new HashMap<>();

        //String hdfsReadToHiveAccessTypes    = rootPlugin.getConfig().get(rootPlugin.getConfig().getPropertyPrefix() + PROP_HDFS_TO_HIVE_ACCESSTYPE_MAPPING_READ, HIVE_ACCESS_TYPE_SELECT + ACCESS_TYPE_SEPARATOR + HIVE_ACCESS_TYPE_REPLADMIN);
        String hdfsReadToHiveAccessTypes    = rootPlugin.getConfig().get(rootPlugin.getConfig().getPropertyPrefix() + PROP_HDFS_TO_HIVE_ACCESSTYPE_MAPPING_READ, HIVE_ACCESS_TYPE_SELECT);
        //String hdfsWriteToHiveAccessTypes   = rootPlugin.getConfig().get(rootPlugin.getConfig().getPropertyPrefix() + PROP_HDFS_TO_HIVE_ACCESSTYPE_MAPPING_WRITE, HIVE_ACCESS_TYPE_UPDATE + ACCESS_TYPE_SEPARATOR + HIVE_ACCESS_TYPE_ALTER + ACCESS_TYPE_SEPARATOR + HIVE_ACCESS_TYPE_DROP + ACCESS_TYPE_SEPARATOR + HIVE_ACCESS_TYPE_REPLADMIN);
        String hdfsWriteToHiveAccessTypes   = rootPlugin.getConfig().get(rootPlugin.getConfig().getPropertyPrefix() + PROP_HDFS_TO_HIVE_ACCESSTYPE_MAPPING_WRITE, HIVE_ACCESS_TYPE_UPDATE + ACCESS_TYPE_SEPARATOR + HIVE_ACCESS_TYPE_ALTER);
        String hdfsExecuteToHiveAccessTypes = rootPlugin.getConfig().get(rootPlugin.getConfig().getPropertyPrefix() + PROP_HDFS_TO_HIVE_ACCESSTYPE_MAPPING_EXECUTE, RangerPolicyEngine.ANY_ACCESS);

        ret.put(HDFS_ACCESS_TYPE_READ,         Arrays.asList(hdfsReadToHiveAccessTypes.split(ACCESS_TYPE_SEPARATOR)));
        ret.put(HDFS_ACCESS_TYPE_WRITE,        Arrays.asList(hdfsWriteToHiveAccessTypes.split(ACCESS_TYPE_SEPARATOR)));
        ret.put(HDFS_ACCESS_TYPE_EXECUTE,      Arrays.asList(hdfsExecuteToHiveAccessTypes.split(ACCESS_TYPE_SEPARATOR)));
        ret.put(RangerPolicyEngine.ANY_ACCESS, Collections.singletonList(RangerPolicyEngine.ANY_ACCESS));

        return ret;
    }

}
